from typing import Optional, List
from decimal import Decimal

from validator_collection import validators, checkers

from highcharts_core import errors
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class AudioParameter(HighchartsMeta):
    """Configuration of an audio parameter's settings."""
    
    def __init__(self, **kwargs):
        self._map_function = None
        self._map_to = None
        self._max = None
        self._min = None
        self._value = None
        self._within = None
        
        self.map_function = kwargs.get('map_function', None)
        self.map_to = kwargs.get('map_to', None)
        self.max = kwargs.get('max', None)
        self.min = kwargs.get('min', None)
        self.value = kwargs.get('value', None)
        self.within = kwargs.get('within', None)
        
    @property
    def map_function(self) -> Optional[str]:
        """The name of the mapping function to apply. Accepts either ``'linear'`` or ``'logarithmic'``.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._map_function
    
    @map_function.setter
    def map_function(self, value):
        if not value:
            self._map_function = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['linear', 'logarithmic']:
                raise errors.HighchartsValueError(f'.map_function expects either "linear" or "logarithmic". '
                                                  f'Received: "{value}"')
            self._map_function = value

    @property
    def map_to(self) -> Optional[str]:
        """The name of the point property to map to determine the audio setting.
        
        .. tip::
        
          A negative sign (``'-'``) can be placed *before* the name of the property to invert
          the mapping.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._map_to
    
    @map_to.setter
    def map_to(self, value):
        self._map_to = validators.string(value, allow_empty = True)
    
    @property
    def max(self) -> Optional[int | float | Decimal]:
        """The maximum value for the audio parameter. This is the highest value the audio parameter will be 
        mapped to.
        
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max
    
    @max.setter
    def max(self, value):
        self._max = validators.numeric(value, allow_empty = True)
        
    @property
    def min(self) -> Optional[int | float | Decimal]:
        """The minimum value for the audio parameter. This is the lowest value the audio parameter will be 
        mapped to.
        
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min
    
    @min.setter
    def min(self, value):
        self._min = validators.numeric(value, allow_empty = True)
        
    @property
    def value(self) -> Optional[int | float | Decimal]:
        """A fixed value to use in place of the mapped ``prop`` when mapping.
        
        .. note::
        
          For example, if mapping to ``'y'``, setting ``.value = 4`` will map as if all
          points had a ``y`` value equal of ``4``.
          
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._value
    
    @value.setter
    def value(self, value):
        self._value = validators.numeric(value, allow_empty = True)
        
    @property
    def within(self) -> Optional[str]:
        """The data values within which to map the audio parameter. Accepts ``'series'``, ``'chart'``, ``'yAxis'``, 
        ``'xAxis'``. Defaults to :obj:`None <python:None>`.
        
        .. note::
        
          Mapping within ``'series'`` will make the lowest value point in the *series* map to the 
          :meth:`.min <highcharts_core.options.sonification.mapping.AudioParameter.min>` audio parameter value, and the 
          highest value will map to the :meth:`.max <highcharts_core.options.sonification.mapping.AudioParameter.max>` 
          audio parameter.
          
          Mapping within ``'chart'`` will make the lowest value point in the *chart* (across all series) map to the 
          :meth:`.min <highcharts_core.options.sonification.mapping.AudioParameter.min>` audio parameter value, and the 
          highest value will map to the :meth:`.max <highcharts_core.options.sonification.mapping.AudioParameter.max>` 
          audio parameter.
          
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._within

    @within.setter
    def within(self, value):
        if not value:
            self._within = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['series', 'chart', 'xaxis', 'yaxis']:
                raise errors.HighchartsValueError(f'.within expects a value of either "series", "chart", '
                                                  f'"xAxis", or "yAxis". Received: "{value}"')
            if value == 'xaxis':
                value = 'xAxis'
            elif value == 'yaxis':
                value = 'yAxis'
                
            self._within = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'map_function': as_dict.get('mapFunction', None),
            'map_to': as_dict.get('mapTo', None),
            'max': as_dict.get('max', None),
            'min': as_dict.get('min', None),
            'value': as_dict.get('value', None),
            'within': as_dict.get('within', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'mapFunction': self.map_function,
            'mapTo': self.map_to,
            'max': self.max,
            'min': self.min,
            'value': self.value,
            'within': self.within,
        }

        return untrimmed


class PitchParameter(AudioParameter):
    """The configuration settings to configure :term:`pitch`.
    
    By default, this object maps to ``'y'``, so low ``'y'`` values are played with a lower pitch
    and high ``'y'`` values are played with a higher pitch.
    
    .. note::
    
      The properties for configuring pitch accept :class:`str <python:str>` values as well as 
      numerical values, specifically allowing you to indicate the musical note that you wish the
      audio track to play. These notes are expressed in the form ``<note><octave>``, such as ``'c6'`` or
      ``'F#2'``.
      
    .. tip::
    
      You configure the pitch to map to an array of notes, with the delay between notes determined by the
      :meth:`SonificationMapping.gap_between_notes <highcharts_core.options.sonification.mapping.SonificationMapping.gap_between_notes>`
      property.
      
    .. tip::
    
      You can also define a musical :meth:`.scale <highcharts_core.options.sonification.mapping.PitchParameter.scale>`
      to follow when mapping.
    
    """
    
    def __init__(self, **kwargs):
        self._scale = None
        
        self.scale = kwargs.get('scale', None)
        
        super().__init__(**kwargs)

    @property
    def map_to(self) -> Optional[str]:
        """The name of the point property to map to determine the audio setting. Defaults to ``'y'``.
        
        .. tip::
        
          A negative sign (``'-'``) can be placed *before* the name of the property to invert
          the mapping.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._map_to
    
    @map_to.setter
    def map_to(self, value):
        self._map_to = validators.string(value, allow_empty = True)
    
    @property
    def max(self) -> Optional[str | int | float | Decimal]:
        """The maximum value for the audio parameter. This is the highest value the audio parameter will be 
        mapped to. Defaults to ``'c6'``.
        
        :rtype: numeric, :class:`str <python:str>`, or :obj:`None <python:None>`
        """
        return self._max
    
    @max.setter
    def max(self, value):
        if value is None:
            self._max = None
        else:
            try:
                self._max = validators.numeric(value)
            except (ValueError, TypeError):
                self._max = validators.string(value)
        
    @property
    def min(self) -> Optional[str | int | float | Decimal]:
        """The minimum value for the audio parameter. This is the lowest value the audio parameter will be 
        mapped to. Defaults to ``'c2'``.
        
        :rtype: numeric, :class:`str <python:str>`, or :obj:`None <python:None>`
        """
        return self._min
    
    @min.setter
    def min(self, value):
        if value is None:
            self._min = None
        else:
            try:
                self._min = validators.numeric(value)
            except (ValueError, TypeError):
                self._min = validators.string(value)
        
    @property
    def scale(self) -> Optional[List[str | int | float | Decimal]]:
        """The scale to map pitches against, defined as an array of semitone offsets from the root note.
        
        :rtype: :class:`list <python:list>` of :class:`str <python:str>` in the form ``'<note><octave>'`` 
          (e.g. ``'c2'``, ``'F#2'``, etc.) or numeric values, or :obj:`None <python:None>`
        """
        return self._scale
    
    @scale.setter
    def scale(self, value):
        if not value:
            self._scale = None
        else:
            value = validators.iterable(value, forbid_literals = (str, bytes, dict))
            validated = []
            for item in value:
                try:
                    item = validators.numeric(item)
                except (ValueError, TypeError):
                    item = validators.string(item)
                    
                validated.append(item)

            self._scale = [x for x in validated]

    @property
    def value(self) -> Optional[int | float | Decimal | str]:
        """A fixed value to use in place of the mapped ``prop`` when mapping. Defaults to
        :obj:`None <python:None>`.
        
        .. note::
        
          For example, if mapping to ``'y'``, setting ``.value = 4`` will map as if all
          points had a ``y`` value equal of ``4``.
          
        :rtype: :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._value
    
    @value.setter
    def value(self, value):
        if value is None:
            self._value = None
        else:
            try:
                self._value = validators.numeric(value)
            except (ValueError, TypeError):
                self._value = validators.string(value)
        
    @property
    def within(self) -> Optional[str]:
        """The data values within which to map the audio parameter. Accepts ``'series'``, ``'chart'``, ``'yAxis'``, 
        ``'xAxis'``. Defaults to ``'yAxis'``.
        
        .. note::
        
          Mapping within ``'series'`` will make the lowest value point in the *series* map to the 
          :meth:`.min <highcharts_core.options.sonification.mapping.AudioParameter.min>` audio parameter value, and the 
          highest value will map to the :meth:`.max <highcharts_core.options.sonification.mapping.AudioParameter.max>` 
          audio parameter.
          
          Mapping within ``'chart'`` will make the lowest value point in the *chart* (across all series) map to the 
          :meth:`.min <highcharts_core.options.sonification.mapping.AudioParameter.min>` audio parameter value, and the 
          highest value will map to the :meth:`.max <highcharts_core.options.sonification.mapping.AudioParameter.max>` 
          audio parameter.
          
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._within

    @within.setter
    def within(self, value):
        if not value:
            self._within = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['series', 'chart', 'xaxis', 'yaxis']:
                raise errors.HighchartsValueError(f'.within expects a value of either "series", "chart", '
                                                  f'"xAxis", or "yAxis". Received: "{value}"')
            if value == 'xaxis':
                value = 'xAxis'
            elif value == 'yaxis':
                value = 'yAxis'
                
            self._within = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'map_function': as_dict.get('mapFunction', None),
            'map_to': as_dict.get('mapTo', None),
            'max': as_dict.get('max', None),
            'min': as_dict.get('min', None),
            'scale': as_dict.get('scale', None),
            'value': as_dict.get('value', None),
            'within': as_dict.get('within', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'mapFunction': self.map_function,
            'mapTo': self.map_to,
            'max': self.max,
            'min': self.min,
            'scale': self.scale,
            'value': self.value,
            'within': self.within,
        }

        return untrimmed


class AudioFilter(HighchartsMeta):
    """Configuration of an audio filter, such as a :term:`highpass` or :term:`lowpass` filter.
    """

    def __init__(self, **kwargs):
        self._frequency = None
        self._resonance = None
        
        self.frequency = kwargs.get('frequency', None)
        self.resonance = kwargs.get('resonance', None)

    @property
    def frequency(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """The frequency, expressed in Hertz, between 1 to 20,000 Hz.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._frequency
    
    @frequency.setter
    def frequency(self, value):
        if value is None:
            self._frequency = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value,
                                                   minimum = 1,
                                                   maximum = 20000)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._frequency = value

    @property
    def resonance(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """The resonance, expressed in Db. Can be negative to cause a dip, or positive to cause a bump.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._resonance
    
    @resonance.setter
    def resonance(self, value):
        if value is None:
            self._resonance = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._resonance = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'frequency': as_dict.get('frequency', None),
            'resonance': as_dict.get('resonance', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'frequency': self.frequency,
            'resonance': self.resonance,
        }

        return untrimmed


class TremoloEffect(HighchartsMeta):
    """Configuration of a :term:`tremolo` effect."""
    
    def __init__(self, **kwargs):
        self._depth = None
        self._speed = None
        
        self.depth = kwargs.get('depth', None)
        self.speed = kwargs.get('speed', None)

    @property
    def depth(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """The intensity of the :term:`tremolo` effect over time, mapped from ``0`` to ``1``.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._depth
    
    @depth.setter
    def depth(self, value):
        if value is None:
            self._depth = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value,
                                                   minimum = 0,
                                                   maximum = 1)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._depth = value

    @property
    def speed(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """The speed of the :term:`tremolo` effect, mapped from ``0`` to ``1``, which determines how rapidly
        the volume changes.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._speed

    @speed.setter
    def speed(self, value):
        if value is None:
            self._speed = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value,
                                                   minimum = 0,
                                                   maximum = 1)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._speed = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'depth': as_dict.get('depth', None),
            'speed': as_dict.get('speed', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'depth': self.depth,
            'speed': self.speed,
        }

        return untrimmed


class SonificationMapping(HighchartsMeta):
    """Mapping options for audio parameters.
    
    All properties in this object accept either:
    
      * an object instance (or coercable to an object instance) as outlined in the documentation
      * a :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>`
      * a :class:`str <python:str>`
      * a numeric value
      
    If supplying a :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>`,
    the function is expected to:
      
      * be called for each audio event to be played
      * receive a context object parameter containing ``time``, and potentially ``point`` and ``value`` depending on the
        track being played (``point`` provided if the audio event is related to a data point, and ``value`` if the 
        track is used as a context track with 
        :meth:`.value_interval <highcharts_core.options.sonification.track_configurations.ContextTrackConfiguration.value_interval>` set).
        
    """
    
    def __init__(self, **kwargs):
        self._frequency = None
        self._gap_between_notes = None
        self._highpass = None
        self._lowpass = None
        self._note_duration = None
        self._pan = None
        self._pitch = None
        self._play_delay = None
        self._rate = None
        self._text = None
        self._time = None
        self._tremolo = None
        self._volume = None
        
        self.frequency = kwargs.get('frequency', None)
        self.gap_between_notes = kwargs.get('gap_between_notes', None)
        self.highpass = kwargs.get('highpass', None)
        self.lowpass = kwargs.get('lowpass', None)
        self.note_duration = kwargs.get('note_duration', None)
        self.pan = kwargs.get('pan', None)
        self.pitch = kwargs.get('pitch', None)
        self.play_delay = kwargs.get('play_delay', None)
        self.rate = kwargs.get('rate', None)
        self.text = kwargs.get('text', None)
        self.time = kwargs.get('time', None)
        self.tremolo = kwargs.get('tremolo', None)
        self.volume = kwargs.get('volume', None)
        
    @property
    def frequency(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """The frequency, expressed in Hertz, of notes. 
        
        Overrides :meth:`.pitch <highcharts_core.options.sonification.mapping.SonificationMapping.pitch>` if set.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._frequency
    
    @frequency.setter
    def frequency(self, value):
        if value is None:
            self._frequency = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._frequency = value
    
    @property
    def gap_between_notes(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """The gap in milliseconds between notes if 
        :meth:`.pitch <highcharts_core.options.sonification.mapping.SonificationMapping.pitch>` is mapped to an
        array of notes.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._gap_between_notes
    
    @gap_between_notes.setter
    def gap_between_notes(self, value):
        if value is None:
            self._gap_between_notes = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._gap_between_notes = value

    @property
    def highpass(self) -> Optional[AudioFilter]:
        """Configuration for the :term:`highpass` filter.
        
        :rtype: :class:`AudioFilter <highcharts_core.options.sonification.mapping.AudioFilter>` or 
          :obj:`None <python:None>`
        """
        return self._highpass
    
    @highpass.setter
    @class_sensitive(AudioFilter)
    def highpass(self, value):
        self._highpass = value
        
    @property
    def lowpass(self) -> Optional[AudioFilter]:
        """Configuration for the :term:`lowpass` filter.
        
        :rtype: :class:`AudioFilter <highcharts_core.options.sonification.mapping.AudioFilter>` or 
          :obj:`None <python:None>`
        """
        return self._lowpass
    
    @lowpass.setter
    @class_sensitive(AudioFilter)
    def lowpass(self, value):
        self._lowpass = value

    @property
    def note_duration(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """How long a note plays in a sustained fashion, expressed in milliseconds.
        
        .. note::
        
          It only affects instruments that are able to play continuous sustained notes. Examples of these instruments 
          from the presets include: ``'flute'``, ``'saxophone'``, ``'trumpet'``, ``'sawsynth'``, ``'wobble'``, 
          ``'basic1'``, ``'basic2'``, ``'sine'``, ``'sineGlide'``, ``'triangle'``, ``'square'``, ``'sawtooth'``, 
          ``'noise'``, ``'filteredNoise'``, and ``'wind'``.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._note_duration
    
    @note_duration.setter
    def note_duration(self, value):
        if value is None:
            self._note_duration = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._note_duration = value
    
    @property
    def pan(self)  -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """The stereo panning position of the sound, defined between left (``-1``) and right (``1``).
        Defaults to a mapping object which maps to the ``'x'`` property, which causes the sound to move 
        from left to right as the chart plays.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._pan
    
    @pan.setter
    def pan(self, value):
        if value is None:
            self._pan = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value,
                                                   minimum = -1,
                                                   maximum = 1)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._pan = value
    
    @property
    def pitch(self) -> Optional[List[PitchParameter | CallbackFunction | str | int | float | Decimal] | PitchParameter | CallbackFunction | str | int | float | Decimal]:
        """Configuration of :term:`pitch` for the audio track.
        
        By default, this object maps to ``'y'``, so low ``'y'`` values are played with a lower pitch
        and high ``'y'`` values are played with a higher pitch.
        
        .. note::
        
          The properties for configuring pitch accept :class:`str <python:str>` values as well as 
          numerical values, specifically allowing you to indicate the musical note that you wish the
          audio track to play. These notes are expressed in the form ``<note><octave>``, such as ``'c6'`` or
          ``'F#2'``.
        
        .. tip::
        
          You configure the pitch to map to an array of notes, with the delay between notes determined by the
          :meth:`SonificationMapping.gap_between_notes <highcharts_core.options.sonification.mapping.SonificationMapping.gap_between_notes>`
          property.
        
        .. tip::
        
          You can also define a musical 
          :meth:`.scale <highcharts_core.options.sonification.mapping.PitchParameter.scale>`
          to follow when mapping.
        
        :rtype: iterable of :class:`PitchParameter <highcharts_core.options.sonification.mapping.PitchParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric, or :class:`PitchParameter <highcharts_core.options.sonification.mapping.PitchParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric, or:obj:`None <python:None>`
        """
        return self._pitch
    
    @pitch.setter
    def pitch(self, value):
        def check_pitch_values(item):
            try:
                item = validate_types(item, types = (PitchParameter))
            except (ValueError, TypeError):
                try:
                    item = validate_types(item, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        item = validators.numeric(item)
                    except (ValueError, TypeError):
                        item = validators.string(item)
            
            return item
            
        if value is None:
            self._pitch = None
        elif checkers.is_iterable(value, forbid_literals = (str, bytes, dict)):
            self._pitch = [check_pitch_values(x) for x in value]
        else:
            self._pitch = check_pitch_values(value)
    
    @property
    def play_delay(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """The number of milliseconds to wait before playing, which comes in addition to the time determined by
        the :meth:`.time <highcharts_core.options.sonification.mapping.SonificationMapping.time>` property.
        
        .. tip::
        
          Can also be negative, which can cause the audio to play *before* the mapped time.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._play_delay
    
    @play_delay.setter
    def play_delay(self, value):
        if value is None:
            self._play_delay = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._play_delay = value
    
    @property
    def rate(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """Mapping configuration for the speech rate multiplier.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._rate
    
    @rate.setter
    def rate(self, value):
        if value is None:
            self._rate = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._rate = value
    
    @property
    def text(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """Mapping configuration for the text parameter for speech tracks.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._text
    
    @text.setter
    def text(self, value):
        if value is None:
            self._text = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._text = value
    
    @property
    def time(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """The offset, expressed in milliseconds, which determines when audio for a point is played (e.g. a value of 
        ``'0'`` means it plays immediately when the chart is sonified).
        
        By default, it is mapped to the ``'x'`` property, which means that points with the lowest ``'x'`` values will 
        play first, while points with the highest value will play last.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._time
    
    @time.setter
    def time(self, value):
        if value is None:
            self._time = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._time = value
    
    @property
    def tremolo(self) -> Optional[TremoloEffect]:
        """Mapping options for a :term:`tremolo` effect.
        
        :rtype: :class:`TremoloEffect <highcharts_core.options.sonification.mapping.TremoloEffect>` or 
          :obj:`None <python:None>`
        """
        return self._tremolo

    @tremolo.setter
    def tremolo(self, value):
        if value is None:
            self._tremolo = None
        else:
            try:
                value = validate_types(value, types = (TremoloEffect))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value)
                    except (ValueError, TypeError):
                        value = validators.string(value)

            self._tremolo = value

    @property
    def volume(self) -> Optional[AudioParameter | CallbackFunction | str | int | float | Decimal]:
        """The volume at which to play notes, expressed from ``0`` (muted) to ``1``.
        
        :rtype: :class:`AudioParameter <highcharts_core.options.sonification.mapping.AudioParameter>` or
          :class:`CallbackFunction <highcharts_core.utility_functions.javascript_functions.CallbackFunction>` or
          :class:`str <python:str>` or numeric or :obj:`None <python:None>`
        """
        return self._volume
    
    @volume.setter
    def volume(self, value):
        if value is None:
            self._volume = None
        else:
            try:
                value = validate_types(value, types = (AudioParameter))
            except (ValueError, TypeError):
                try:
                    value = validate_types(value, types = (CallbackFunction))
                except (ValueError, TypeError):
                    try:
                        value = validators.numeric(value,
                                                   minimum = 0,
                                                   maximum = 1)
                    except (ValueError, TypeError):
                        value = validators.string(value)
                        
            self._volume = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'frequency': as_dict.get('frequency', None),
            'gap_between_notes': as_dict.get('gapBetweenNotes', None),
            'highpass': as_dict.get('highpass', None),
            'lowpass': as_dict.get('lowpass', None),
            'note_duration': as_dict.get('noteDuration', None),
            'pan': as_dict.get('pan', None),
            'pitch': as_dict.get('pitch', None),
            'play_delay': as_dict.get('playDelay', None),
            'rate': as_dict.get('rate', None),
            'text': as_dict.get('text', None),
            'time': as_dict.get('time', None),
            'tremolo': as_dict.get('tremolo', None),
            'volume': as_dict.get('volume', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'frequency': self.frequency,
            'gapBetweenNotes': self.gap_between_notes,
            'highpass': self.highpass,
            'lowpass': self.lowpass,
            'noteDuration': self.note_duration,
            'pan': self.pan,
            'pitch': self.pitch,
            'playDelay': self.play_delay,
            'rate': self.rate,
            'text': self.text,
            'time': self.time,
            'tremolo': self.tremolo,
            'volume': self.volume,
        }

        return untrimmed



from typing import Optional, List

from validator_collection import validators, checkers
import esprima
from esprima.error_handler import Error as ParseError

from highcharts_core import errors
from highcharts_core.decorators import validate_types
from highcharts_core.metaclasses import HighchartsMeta


class CallbackFunction(HighchartsMeta):
    """Representation of a JavaScript callback function's source code."""

    def __init__(self, **kwargs):
        self._function_name = None
        self._arguments = None
        self._body = None

        self.function_name = kwargs.get('function_name', None)
        self.arguments = kwargs.get('arguments', None)
        self.body = kwargs.get('body', None)

    def __str__(self) -> str:
        if self.function_name:
            prefix = f'function {self.function_name}'
        else:
            prefix = 'function'

        arguments = '('
        if self.arguments:
            for argument in self.arguments:
                arguments += f'{argument},'
            arguments = arguments[:-1]

        arguments += ')'

        as_str = f'{prefix}{arguments}'
        as_str += ' {'
        if self.body:
            as_str += '\n'
            as_str += self.body

        as_str += '}'

        return as_str

    @property
    def function_name(self) -> Optional[str]:
        """An optional name to be given to the function.

        .. warning::

          Most Highcharts Callback function definitions are anonymous, meaning that they
          are named within the object into which they are embedded. As a result,
          this setting should be used sparingly.

        :rtype: :class:`str <python:str>`
        """
        return self._function_name

    @function_name.setter
    def function_name(self, value):
        self._function_name = validators.variable_name(value, allow_empty = True)

    @property
    def arguments(self) -> Optional[List[str]]:
        """Collection of named arguments (parameters) that will be passed to the function.

        :rtype: :class:`list <python:list>` of :obj:`str <python:str>`, or
          :obj:`None <python:None>`
        """
        return self._arguments

    @arguments.setter
    def arguments(self, value):
        if not value:
            self._arguments = None
        else:
            self._arguments = [validators.variable_name(x)
                               for x in validators.iterable(value)]

    @property
    def body(self) -> Optional[str]:
        """The source code of the function itself.

        .. note::

          Should *not* be wrapped in ``{ ... }``. It should just be the source code of the
          the function itself.

        .. hint::

          When writing this code in Python, it is best to use the three-quotation-mark
          string pattern, like so:

          .. code-block:: python

            callback = CallbackFunction()
            callback.body = \"\"\"
            ... some JavaScript logic goes here
            \"\"\"

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._body

    @body.setter
    def body(self, value):
        self._body = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'function_name': as_dict.get('function_name',
                                         as_dict.get('functionName', None)),
            'arguments': as_dict.get('arguments', None),
            'body': as_dict.get('body', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'function_name': self.function_name,
            'arguments': self.arguments,
            'body': self.body
        }

    def to_json(self, encoding = 'utf-8'):
        return None

    def to_js_literal(self,
                      filename = None,
                      encoding = 'utf-8') -> str:
        if filename:
            filename = validators.path(filename)

        as_str = str(self)

        if filename:
            with open(filename, 'w', encoding = encoding) as file_:
                file_.write(as_str)

        return as_str

    @classmethod
    def _convert_from_js_ast(cls, property_definition, original_str):
        """Create a :class:`CallbackFunction` instance from a
        :class:`esprima.nodes.FunctionExpression` instance.

        :param property_definition: The :class:`esprima.nodes.FunctionExpression`
          instance, including ``loc`` (indicating the line and column in the original
          string) and ``range`` (indicating the character range for the property
          definition in the original string).
        :type property_definition: :class:`esprima.nodes.FunctionExpression`

        :param original_str: The original :class:`str <python:str>` of the JavaScript from
          which ``property_definition`` was parsed.
        :type original_str: :class:`str <python:str>`

        :returns: :class:`CallbackFunction`
        """
        if not checkers.is_type(property_definition, ('FunctionDeclaration',
                                                      'FunctionExpression',
                                                      'MethodDefinition',
                                                      'Property')):
            raise errors.HighchartsParseError(f'property_definition should contain a '
                                              f'FunctionExpression, FunctionDeclaration, '
                                              'MethodDefinition, or Property instance. '
                                              f'Received: '
                                              f'{property_definition.__class__.__name__}')

        if property_definition.type not in ['MethodDefinition', 'Property']:
            body = property_definition.body
        else:
            body = property_definition.value.body

        body_range = body.range
        body_start = body_range[0] + 1
        body_end = body_range[1] - 1

        if property_definition.type == 'FunctionDeclaration':
            function_name = property_definition.id.name
        elif property_definition.type == 'MethodDefinition':
            function_name = property_definition.key.name
        elif property_definition.type == 'FunctionExpression' and \
             property_definition.id is not None:
            function_name = property_definition.id.name
        else:
            function_name = None

        function_body = original_str[body_start:body_end]

        if property_definition.type in ['MethodDefinition', 'Property']:
            arguments = [x.name for x in property_definition.value.params]
        else:
            arguments = [x.name for x in property_definition.params]

        return cls(function_name = function_name,
                   arguments = arguments,
                   body = function_body)

    @classmethod
    def from_js_literal(cls,
                        as_str_or_file,
                        allow_snake_case: bool = True,
                        _break_loop_on_failure: bool = False):
        """Return a Python object representation of a Highcharts JavaScript object
        literal.

        :param as_str_or_file: The JavaScript object literal, represented either as a
          :class:`str <python:str>` or as a filename which contains the JS object literal.
        :type as_str_or_file: :class:`str <python:str>`

        :param allow_snake_case: If ``True``, interprets ``snake_case`` keys as equivalent
          to ``camelCase`` keys. Defaults to ``True``.
        :type allow_snake_case: :class:`bool <python:bool>`

        :param _break_loop_on_failure: If ``True``, will break any looping operations in
          the event of a failure. Otherwise, will attempt to repair the failure. Defaults
          to ``False``.
        :type _break_loop_on_failure: :class:`bool <python:bool>`

        :returns: A Python object representation of the Highcharts JavaScript object
          literal.
        :rtype: :class:`HighchartsMeta`
        """
        is_file = checkers.is_file(as_str_or_file)
        if is_file:
            with open(as_str_or_file, 'r') as file_:
                as_str = file_.read()
        else:
            as_str = as_str_or_file

        parsed, updated_str = cls._validate_js_function(as_str)
        if parsed.body[0].type == 'FunctionDeclaration':
            property_definition = parsed.body[0]
        elif parsed.body[0].type == 'MethodDefinition':
            property_definition = parsed.body[0].body[0]
        elif parsed.body[0].type != 'FunctionDeclaration':
            property_definition = parsed.body[0].declarations[0].init

        return cls._convert_from_js_ast(property_definition, updated_str)

    @classmethod
    def _validate_js_function(cls,
                              as_str,
                              range = True,
                              _break_loop_on_failure = False):
        """Parse a JavaScript function from within ``as_str``.

        :param as_str: A string that potentially contains a JavaScript function.
        :rtype: :class:`str <python:str>`

        :param range: If ``True``, include each node's ``loc`` and ``range`` in the AST
          produced. Defaults to ``True``.
        :type range: :class:`bool <python:bool>`

        :param _break_loop_on_failure: If ``True``, prevents

        :returns: 2-member tuple, with the first being a parsed AST of the function and
          the second being the string that ultimatley produced that parsed AST.
        :rtype: :class:`tuple <python:tuple>` of :class:`esprima.nodes.Script`,
          :class:`str <python:str>`
        """
        try:
            parsed = esprima.parseScript(as_str, loc = range, range = range)
        except ParseError:
            try:
                parsed = esprima.parseModule(as_str, loc = range, range = range)
            except ParseError:
                if not _break_loop_on_failure and as_str.startswith('function'):
                    as_str = f"""const testFunction = {as_str}"""
                    return cls._validate_js_function(as_str,
                                                     range = range,
                                                     _break_loop_on_failure = True)
                elif not _break_loop_on_failure:
                    as_str = f"""const testFunction = function {as_str}"""
                    return cls._validate_js_function(as_str,
                                                     range = range,
                                                     _break_loop_on_failure = True)
                else:
                    raise errors.HighchartsParseError('._validate_js_function() expects '
                                                      'a str containing a valid '
                                                      'JavaScript function. Could not '
                                                      'find a valid function.')

        return parsed, as_str


class JavaScriptClass(HighchartsMeta):
    """Representation of a JavaScript class."""

    def __init__(self, **kwargs):
        self._class_name = None
        self._methods = None

        self.class_name = kwargs.get('class_name', None)
        self.methods = kwargs.get('methods', None)

    def __str__(self) -> str:
        if not self.class_name:
            raise errors.HighchartsMissingClassNameError('Unable to serialize. The '
                                                         'JavaScriptClass instance has '
                                                         'no class_name provided.')
        as_str = f'class {self.class_name} '
        as_str += '{\n'
        for method in self.methods or []:
            method_str = f'{method.function_name}'
            argument_str = '('
            for argument in method.arguments or []:
                argument_str += f'{argument},'
            if method.arguments:
                argument_str = argument_str[:-1]
            argument_str += ') {\n'

            method_str += argument_str

            method_str += method.body + '\n}\n'

            as_str += method_str

        as_str += '}'

        return as_str

    @property
    def class_name(self) -> Optional[str]:
        """The name of the JavaScript class.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.variable_name(value, allow_empty = True)

    @property
    def methods(self) -> Optional[List[CallbackFunction]]:
        """Collection of methods that are to be defined within the class. Defaults to
        :obj:`None <python:None>`.

        .. warning::

          All methods *must* have a :meth:`function_name <CallbackFunction.function_name>`
          set.

        .. warning::

          One of the methods *must* have a
          :meth:`function_name <CallbackFunction.function_name>` of ``'constructor'`` and
          be used as a constructor for the class.

        .. note::

          For the sake of simplicity, the :class:`JavaScriptClass` does not support
          ECMAScript's more robust public/private field declaration syntax, nor does it
          support the definition of getters or generators.

        :rtype: :class:`list <python:list>` of :class:`CallbackFunction`, or
          :obj:`None <python:None>`

        :raises HighchartsJavaScriptError: if one or more methods lacks a function name OR
          if there is no ``constructor`` method included in
          :meth:`.methods <JavaScriptClass.methods>`.
        """
        return self._methods

    @methods.setter
    def methods(self, value):
        if not value:
            self._methods = None
        else:
            value = validate_types(value,
                                   types = CallbackFunction,
                                   force_iterable = True)
            has_constructor = False
            for method in value:
                if not method.function_name:
                    raise errors.HighchartsJavaScriptError('All JavaScriptClass methods '
                                                           'require a function name.')
                if method.function_name == 'constructor':
                    has_constructor = True

            if not has_constructor:
                raise errors.HighchartsJavaScriptError('A JavaScriptClass requires at '
                                                       'least one "constructor" method. '
                                                       'Yours had none.')

            self._methods = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'class_name': as_dict.get('className', None),
            'methods': as_dict.get('methods', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'className': self.class_name,
            'methods': self.methods
        }

    @classmethod
    def _convert_from_js_ast(cls, definition, original_str):
        """Create a :class:`JavaScriptClass` instance from a
        :class:`esprima.nodes.ClassDeclaration` instance.

        :param property_definition: The :class:`esprima.nodes.ClassDeclaration` instance,
          including ``loc`` (indicating the line and column in the original string) and
          ``range`` (indicating the character range for the property definition in the
          original string).
        :type property_definition: :class:`esprima.nodes.ClassDeclaration`

        :param original_str: The original :class:`str <python:str>` of the JavaScript from
          which ``definition`` was parsed.
        :type original_str: :class:`str <python:str>`

        :returns: :class:`JavaScriptClass`
        """
        if not checkers.is_type(definition, ('ClassDeclaration', 'ClassExpression')):
            raise errors.HighchartsParseError(f'definition should contain a '
                                              f'ClassDeclaration or ClassExpression'
                                              ' instance. Received: '
                                              f'{definition.__class__.__name__}')

        class_name = definition.id.name

        method_definitions = [x for x in definition.body.body]
        method_strings = []
        for method in method_definitions:
            method_start = method.range[0]
            method_end = method.range[1]
            method_string = original_str[method_start:method_end]
            method_strings.append(method_string)

        methods = [CallbackFunction.from_js_literal(x) for x in method_strings]

        return cls(class_name = class_name,
                   methods = methods)

    @classmethod
    def from_js_literal(cls,
                        as_str_or_file):
        """Return a Python object representation of a JavaScript class.

        :param as_str_or_file: The JavaScript object literal, represented either as a
          :class:`str <python:str>` or as a filename which contains the JS object literal.
        :type as_str_or_file: :class:`str <python:str>`

        :param _break_loop_on_failure: If ``True``, will break any looping operations in
          the event of a failure. Otherwise, will attempt to repair the failure. Defaults
          to ``False``.
        :type _break_loop_on_failure: :class:`bool <python:bool>`

        :returns: A Python object representation of the Highcharts JavaScript object
          literal.
        :rtype: :class:`HighchartsMeta`
        """
        is_file = checkers.is_file(as_str_or_file)
        if is_file:
            with open(as_str_or_file, 'r') as file_:
                as_str = file_.read()
        else:
            as_str = as_str_or_file

        try:
            parsed = esprima.parseScript(as_str, range = True)
        except ParseError:
            try:
                parsed = esprima.parseModule(as_str, range = True)
            except ParseError:
                raise errors.HighchartsParseError('unable to find a JavaScript class '
                                                  'declaration in ``as_str``.')

        definition = parsed.body[0]

        return cls._convert_from_js_ast(definition, as_str)

    def to_js_literal(self,
                      filename = None,
                      encoding = 'utf-8') -> str:
        if filename:
            filename = validators.path(filename)

        as_str = str(self)

        if filename:
            with open(filename, 'w', encoding = encoding) as file_:
                file_.write(as_str)

        return as_str


class VariableName(HighchartsMeta):
    """Object that represents a (JavaScript) variable name that may be referenced in
    **Highcharts for Python** items."""

    def __init__(self, **kwargs):
        self._variable_name = None

        self.variable_name = kwargs.get('variable_name', None)

    @property
    def variable_name(self) -> Optional[str]:
        """The name of the (JavaScript) variable which will be incorporated into
        serializations of **Highcharts for Python** objects as needed.

        :rtype: :class:`str <python:str>`
        """
        return self._variable_name

    @variable_name.setter
    def variable_name(self, value):
        self._variable_name = validators.variable_name(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'variable_name': as_dict.get('variable_name', as_dict.get('variableName',
                                                                      None)),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'variableName': self.variable_name,
        }

        return untrimmed



from typing import Optional, List

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta

from highcharts_core.options.annotations.animation import AnnotationAnimation
from highcharts_core.options.annotations.control_point_options import AnnotationControlPointOption
from highcharts_core.options.annotations.events import AnnotationEvent
from highcharts_core.options.annotations.label_options import LabelOptions
from highcharts_core.options.annotations.label_options import AnnotationLabel
from highcharts_core.options.annotations.shape_options import ShapeOptions
from highcharts_core.options.annotations.shape_options import AnnotationShape


class Annotation(HighchartsMeta):
    """A basic type of an annotation. It allows adding custom labels or shapes. The
    items can be tied to points, axis coordinates or chart pixel coordinates."""

    def __init__(self, **kwargs):
        self._animation = None
        self._control_point_options = None
        self._crop = None
        self._draggable = None
        self._events = None
        self._id = None
        self._label_options = None
        self._labels = None
        self._shape_options = None
        self._shapes = None
        self._visible = None
        self._z_index = None

        self.animation = kwargs.get('animation', None)
        self.control_point_options = kwargs.get('control_point_options', None)
        self.crop = kwargs.get('crop', None)
        self.draggable = kwargs.get('draggable', None)
        self.events = kwargs.get('events', None)
        self.id = kwargs.get('id', None)
        self.label_options = kwargs.get('label_options', None)
        self.labels = kwargs.get('labels', None)
        self.shape_options = kwargs.get('shape_options', None)
        self.shapes = kwargs.get('shapes', None)
        self.visible = kwargs.get('visible', None)
        self.z_index = kwargs.get('z_index', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'annotations'

    @property
    def animation(self) -> Optional[AnnotationAnimation]:
        """Enable or disable the initial animation when a series is displayed for the
        annotation. If not :obj:`None <python:None>`, is enabled. Otherwise, disabled.

        .. warning::

          This option only applies to the initial animation.

          For other animations, see ``chart.animation`` and the animation parameter under
          the API methods.

        :returns: The configuration settings for the annotation animation.
        :rtype: :class:`AnnotationAnimation`
        """
        return self._animation

    @animation.setter
    @class_sensitive(AnnotationAnimation)
    def animation(self, value):
        self._animation = value

    @property
    def control_point_options(self) -> Optional[AnnotationControlPointOption]:
        """Options for annotation's control points.

        Each control point inherits options from this property, though the global
        options can be overwritten by options in a specific control point.

        :returns: Options configuring the annotations' control points.
        :rtype: :class:`AnnotationControlPointOption` or :obj:`None <python:None>`

        """
        return self._control_point_options

    @control_point_options.setter
    @class_sensitive(AnnotationControlPointOption)
    def control_point_options(self, value):
        self._control_point_options = value

    @property
    def crop(self) -> Optional[bool]:
        """If ``True``, hide the part of the annotation that is outside the plot area.
        Defaults to ``True``.

        :returns: Flag indicating whether to clip an annotation that extends beyond the
          plot area.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._crop

    @crop.setter
    def crop(self, value):
        if value is None:
            self._crop = None
        else:
            self._crop = bool(value)

    @property
    def draggable(self) -> Optional[str]:
        """Setting that allows an annotation to be draggable by a user. Defaults to
        ``'xy'``

        Supports values:

          * ``'x'``
          * ``'xy'``
          * ``'y'``
          * ``''`` (empty string - disables dragging)

        :returns: Configuration of annotation dragging by the user.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises HighchartsValueError: if an unsupported value is supplied
        """
        return self._draggable

    @draggable.setter
    def draggable(self, value):
        if value is None:
            self._draggable = None
        else:
            value = validators.string(value, allow_empty = True) or ''
            value = value.lower()
            if value not in ['x', 'xy', 'y', '']:
                raise errors.HighchartsValueError(f'draggable must be "x", "xy", "y", '
                                                  f'or "". Was: {value}')
            self._draggable = value

    @property
    def events(self) -> Optional[AnnotationEvent]:
        """JavaScript callback functions that fire in response to annotation-related
        events.

        :returns: Callback functions that fire in response to annotation-related events.
        :rtype: :class:`AnnotationEvent` or :obj:`None <python:None>`

        """
        return self._events

    @events.setter
    @class_sensitive(AnnotationEvent)
    def events(self, value):
        self._events = value

    @property
    def id(self) -> Optional[str]:
        """Sets an ID for the annotation. Can be user later when removing an annotation
        using the JavaScript ``Chart.removeAnnotation(id)`` method.

        :returns: The ID for the annotation.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._id

    @id.setter
    def id(self, value):
        self._id = validators.string(value, allow_empty = True)

    @property
    def label_options(self) -> Optional[LabelOptions]:
        """Global options applied to all annotation labels.

        .. note::

          An option from the :meth:`Annotation.label_options` can be overwritten by the
          configuration for a specific label.

        :returns: Configuration options for annotation labels.
        :rtype: :class:`LabelOptions` or :obj:`None <python:None>`
        """
        return self._label_options

    @label_options.setter
    @class_sensitive(LabelOptions)
    def label_options(self, value):
        self._label_options = value

    @property
    def labels(self) -> Optional[List[AnnotationLabel]]:
        """An array of labels to display for annotations.

        .. seealso::

          * :meth:`Annotation.label_options`
          * :class:`LabelOptions`

        :returns: A collection of labels to display for annotation.
        :rtype: :class:`list <python:list>` of :class:`AnnotationLabel` or
          :obj:`None <python:None>`
        """
        return self._labels

    @labels.setter
    @class_sensitive(AnnotationLabel, force_iterable = True)
    def labels(self, value):
        self._labels = value

    @property
    def shape_options(self) -> Optional[ShapeOptions]:
        """Global options applied to all annotation shapes.

        .. note::

          An option from the :meth:`Annotation.shape_options` can be overwritten by the
          configuration for a specific shape.

        :returns: Configuration options for annotation shapes.
        :rtype: :class:`ShapeOptions` or :obj:`None <python:None>`
        """
        return self._shape_options

    @shape_options.setter
    @class_sensitive(ShapeOptions)
    def shape_options(self, value):
        self._shape_options = value

    @property
    def shapes(self) -> Optional[List[AnnotationShape]]:
        """An array of shapes to display for annotations.

        .. seealso::

          * :meth:`Annotation.shape_options`
          * :class:`shapeOptions`

        :returns: A collection of shapes to display for annotation.
        :rtype: :class:`list <python:list>` of :class:`AnnotationShape` or
          :obj:`None <python:None>`
        """
        return self._shapes

    @shapes.setter
    @class_sensitive(AnnotationShape, force_iterable = True)
    def shapes(self, value):
        self._shapes = value

    @property
    def visible(self) -> Optional[bool]:
        """If ``True``, indicates the annotation is visible.

        :returns: Flag which indicates whether the annotation is visible or not.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._visible

    @visible.setter
    def visible(self, value):
        if value is None:
            self._visible = None
        else:
            self._visible = bool(value)

    @property
    def z_index(self) -> Optional[int]:
        """The Z-Index for the annotation. Defaults to ``6``.

        :returns: The z-index for the annotation.
        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._z_index

    @z_index.setter
    def z_index(self, value):
        self._z_index = validators.integer(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'animation': as_dict.get('animation', None),
            'control_point_options': as_dict.get('controlPointOptions', None),
            'crop': as_dict.get('crop', None),
            'draggable': as_dict.get('draggable', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_options': as_dict.get('labelOptions', None),
            'labels': as_dict.get('labels', None),
            'shape_options': as_dict.get('shapeOptions', None),
            'shapes': as_dict.get('shapes', None),
            'visible': as_dict.get('visible', None),
            'z_index': as_dict.get('zIndex', None),
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'animation': self.animation,
            'controlPointOptions': self.control_point_options,
            'crop': self.crop,
            'draggable': self.draggable,
            'events': self.events,
            'id': self.id,
            'labelOptions': self.label_options,
            'labels': self.labels,
            'shapeOptions': self.shape_options,
            'shapes': self.shapes,
            'visible': self.visible,
            'zIndex': self.z_index
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class ExportDataLanguageOptions(HighchartsMeta):
    """Language strings used in the exported data table."""

    def __init__(self, **kwargs):
        self._annotation_header = None
        self._category_datetime_header = None
        self._category_header = None

        self.annotation_header = kwargs.get('annotation_header', None)
        self.category_datetime_header = kwargs.get('category_datetime_header', None)
        self.category_header = kwargs.get('category_header', None)

    @property
    def annotation_header(self) -> Optional[str]:
        """The annotation column title. Defaults to
        ``'Annotations'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._annotation_header

    @annotation_header.setter
    def annotation_header(self, value):
        self._annotation_header = validators.string(value, allow_empty = True)

    @property
    def category_datetime_header(self) -> Optional[str]:
        """The category column title when axis type set to "datetime". Defaults to
        ``'DateTime'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._category_datetime_header

    @category_datetime_header.setter
    def category_datetime_header(self, value):
        self._category_datetime_header = validators.string(value, allow_empty = True)

    @property
    def category_header(self) -> Optional[str]:
        """The category column title. Defaults to
        ``'Category'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._category_header

    @category_header.setter
    def category_header(self, value):
        self._category_header = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'annotation_header': as_dict.get('annotationHeader', None),
            'category_datetime_header': as_dict.get('categoryDatetimeHeader', None),
            'category_header': as_dict.get('categoryHeader', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'annotationHeader': self.annotation_header,
            'categoryDatetimeHeader': self.category_datetime_header,
            'categoryHeader': self.category_header
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.position import Position


class ResetZoomButtonOptions(HighchartsMeta):
    """Configuration settings for the button that appears after a selection zoom,
    allowing the user to reset zoom."""

    def __init__(self, **kwargs):
        self._position = None
        self._relative_to = None
        self._theme = None

        self.position = kwargs.get('position', None)
        self.relative_to = kwargs.get('relative_to', None)
        self.theme = kwargs.get('theme', None)

    @property
    def position(self) -> Optional[Position]:
        """The position of the button.

        :rtype: :class:`Position` or :obj:`None <python:None>`
        """
        return self._position

    @position.setter
    @class_sensitive(Position)
    def position(self, value):
        self._position = value

    @property
    def relative_to(self) -> Optional[str]:
        """What frame the button placement should be related to. Defaults to ``'plot'``.

        Accepts:

          * ``'plot'``
          * ``'chart'``
          * ``'plotBox'``
          * ``'spacingBox'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._relative_to

    @relative_to.setter
    def relative_to(self, value):
        if not value:
            self._relative_to = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['plot', 'chart', 'plotBox', 'spacingBox']:
                raise errors.HighchartsValueError(f'relative_to accepts "plot", "chart", '
                                                  f'"plotBox", "spacingBox", or None. '
                                                  f'Received: {value}')
            self._relative_to = value

    @property
    def theme(self) -> Optional[dict]:
        """A collection of attributes for the button.

        The object takes SVG attributes like ``fill``, ``stroke``, ``stroke-width`` or
        ``r``, the border radius.

        The theme also supports ``style``, a collection of CSS properties for the text.
        Equivalent attributes for the hover state are given in JavaScript
        ``theme.states.hover``.

        :rtype: :class:`dict <python:dict>`
        """
        return self._theme

    @theme.setter
    def theme(self, value):
        if not value or isinstance(value, constants.EnforcedNullType):
            self._theme = None
        else:
            self._theme = validators.dict(value, allow_empty = False)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'position': as_dict.get('position', None),
            'relative_to': as_dict.get('relativeTo', None),
            'theme': as_dict.get('theme', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'position': self.position,
            'relativeTo': self.relative_to,
            'theme': self.theme
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.plot_options.series import SeriesOptions
from highcharts_core.utility_classes.jitter import Jitter


class BubbleOptions(SeriesOptions):
    """General options to apply to all Bubble series types.

    A bubble series is a three dimensional series type where each point renders an X,
    Y, and Z value. Each points is drawn as a bubble where the position along the X
    and Y axes mark the X and Y values, and the size of the bubble relates to the Z
    value.

    .. figure:: ../../../_static/bubble-example.png
      :alt: Bubble Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._display_negative = True
        self._jitter = None
        self._max_size = None
        self._min_size = None
        self._size_by = None
        self._size_by_absolute_value = False
        self._z_max = None
        self._z_min = None
        self._z_threshold = None

        self.display_negative = kwargs.get('display_negative', None)
        self.jitter = kwargs.get('jitter', None)
        self.max_size = kwargs.get('max_size', None)
        self.min_size = kwargs.get('min_size', None)
        self.size_by = kwargs.get('size_by', None)
        self.size_by_absolute_value = kwargs.get('size_by_absolute_value', None)
        self.z_max = kwargs.get('z_max', None)
        self.z_min = kwargs.get('z_min', None)
        self.z_threshold = kwargs.get('z_threshold', None)

        super().__init__(**kwargs)

    @property
    def display_negative(self) -> Optional[bool]:
        """If ``True``, display negative sized bubbles.

        The threshold is given by the :meth:`z_threshold <BubbleOptions.z_threshold>`
        setting, and negative bubbles can be visualized by setting
        :meth:`negative_color <BubbleOptions.negative_color>`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._display_negative

    @display_negative.setter
    def display_negative(self, value):
        if value is None:
            self._display_negative = None
        else:
            self._display_negative = bool(value)

    @property
    def jitter(self) -> Optional[Jitter]:
        """Apply a jitter effect for the rendered markers.

        When plotting discrete values, a little random noise may help telling the points
        apart. The jitter setting applies a random displacement of up to n axis units in
        either direction.

        So for example on a horizontal X axis, setting the ``jitter.x`` to ``0.24`` will
        render the point in a random position between 0.24 units to the left and 0.24
        units to the right of the true axis position. On a category axis, setting it to
        ``0.5`` will fill up the bin and make the data appear continuous.

        When rendered on top of a box plot or a column series, a jitter value of 0.24 will
        correspond to the underlying series' default ``group_padding`` and
        ``point_padding`` settings.

        :rtype: :class:`Jitter` or :obj:`None <python:None>`
        """
        return self._jitter

    @jitter.setter
    @class_sensitive(Jitter)
    def jitter(self, value):
        self._jitter = value

    @property
    def max_size(self) -> Optional[str | int | float | Decimal]:
        """Maximum bubble size. Defaults to ``'20%'``.

        If :obj:`None <python:None>`, bubbles will automatically size between the
        :meth:`min_size <BubbleOptions.min_size>` and ``max_size``, to reflect the z value
        of each bubble. Can be either pixels (when no unit is given), or a percentage of
        the smallest one of the plot width and height.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max_size

    @max_size.setter
    def max_size(self, value):
        if value is None:
            self._max_size = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('max_size expects either a number '
                                                      'or a % string. No "%" found.')
            except TypeError:
                self._max_size = validators.numeric(value,
                                                    allow_empty = True,
                                                    minimum = 0)

    @property
    def min_size(self) -> Optional[str | int | float | Decimal]:
        """Minimum bubble size. Defaults to ``8``.

        If :obj:`None <python:None>`, bubbles will automatically size between the
        :meth:`max_size <BubbleOptions.max_size>` and ``min_size``, to reflect the z value
        of each bubble. Can be either pixels (when no unit is given), or a percentage of
        the smallest one of the plot width and height.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min_size

    @min_size.setter
    def min_size(self, value):
        if value is None:
            self._min_size = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('min_size expects either a number '
                                                      'or a % string. No "%" found.')
            except TypeError:
                self._min_size = validators.numeric(value,
                                                    allow_empty = True,
                                                    minimum = 0)

    @property
    def size_by(self) -> Optional[str]:
        """Whether the bubble's value should be represented by the ``'area'`` or the
        ``'width'`` of the bubble. The default, ``'area'``, corresponds best to the human
        perception of the size of each bubble.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._size_by

    @size_by.setter
    def size_by(self, value):
        if not value:
            self._size_by = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['area', 'width']:
                raise errors.HighchartsValueError(f'size_by expects either "area", '
                                                  f'or "width". Received: {value}')
            self._size_by = value

    @property
    def size_by_absolute_value(self) -> Optional[bool]:
        """If ``True``, the absolute value of z determines the size of the bubble. This
        means that with the default :meth:`z_threshold <BubbleOptions.z_threshold>` of
        ``0``, a bubble of value ``-1`` will have the same size as a bubble of value
        ``1``, while a bubble of value ``0`` will have a smaller size according to
        :meth:`min_size <BubbleOptions.min_size>`.

        Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._size_by_absolute_value

    @size_by_absolute_value.setter
    def size_by_absolute_value(self, value):
        if value is None:
            self._size_by_absolute_value = None
        else:
            self._size_by_absolute_value = bool(value)

    @property
    def z_max(self) -> Optional[int | float | Decimal]:
        """The maximum for the Z value range. When :obj:`None <python:None>`, defaults to
        the highest Z value in the data. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._z_max

    @z_max.setter
    def z_max(self, value):
        self._z_max = validators.numeric(value, allow_empty = True)

    @property
    def z_min(self) -> Optional[int | float | Decimal]:
        """The minimum for the Z value range. When :obj:`None <python:None>`, defaults to
        the lowest Z value in the data. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._z_min

    @z_min.setter
    def z_min(self, value):
        self._z_min = validators.numeric(value, allow_empty = True)

    @property
    def z_threshold(self) -> Optional[int | float | Decimal]:
        """When :meth:`display_negative <BubbleOptions.display_negative>` is ``False``,
        then bubbles with a Z value lower than ``z_threshold`` are not rendered. When
        :meth:`display_negative <BubbleOptions.display_negative>` is ``True`` and
        :meth:`negative_color <BubbleOptions.negative_color>` is set, then points
        with a Z value lower than ``z_threshold`` are rendered with the negative coloring.

        Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._z_threshold

    @z_threshold.setter
    def z_threshold(self, value):
        self._z_threshold = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'point_description_format': as_dict.get('pointDescriptionFormat', None),
            'point_interval': as_dict.get('pointInterval', None),            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'display_negative': as_dict.get('displayNegative', None),
            'jitter': as_dict.get('jitter', None),
            'max_size': as_dict.get('maxSize', None),
            'min_size': as_dict.get('minSize', None),
            'size_by': as_dict.get('sizeBy', None),
            'size_by_absolute_value': as_dict.get('sizeByAbsoluteValue', None),
            'z_max': as_dict.get('zMax', None),
            'z_min': as_dict.get('zMin', None),
            'z_threshold': as_dict.get('zThreshold', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'displayNegative': self.display_negative,
            'jitter': self.jitter,
            'maxSize': self.max_size,
            'minSize': self.min_size,
            'sizeBy': self.size_by,
            'sizeByAbsoluteValue': self.size_by_absolute_value,
            'zMax': self.z_max,
            'zMin': self.z_min,
            'zThreshold': self.z_threshold
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.position import Position


class CreditStyleOptions(HighchartsMeta):
    """CSS styles that are applicable to the :class:`Credits` label."""

    def __init__(self, **kwargs):
        self._color = None
        self._cursor = None
        self._font_size = None

        self.color = kwargs.get('color', None)
        self.cursor = kwargs.get('cursor', None)
        self.font_size = kwargs.get('font_size', None)

    @property
    def color(self) -> Optional[str]:
        """The color to apply to the label. Defaults to ``'#999999'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`.
        """
        return self._color

    @color.setter
    def color(self, value):
        self._color = validators.string(value, allow_empty = True)

    @property
    def cursor(self) -> Optional[str]:
        """The value to pass for the mouse cursor. Defaults to ``'pointer'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._cursor

    @cursor.setter
    def cursor(self, value):
        self._cursor = validators.string(value, allow_empty = True)

    @property
    def font_size(self) -> Optional[str]:
        """The font size to apply to the label. Defaults to ``'9px'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._font_size

    @font_size.setter
    def font_size(self, value):
        self._font_size = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'font_size': as_dict.get('fontSize', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'color': self.color,
            'cursor': self.cursor,
            'fontSize': self.font_size
        }

        return untrimmed


class Credits(HighchartsMeta):
    """Highchart by default puts a credits label in the lower right corner of the
    chart. This can be changed using these options."""

    def __init__(self, **kwargs):
        self._enabled = None
        self._href = None
        self._position = None
        self._style = None
        self._text = None

        self.enabled = kwargs.get('enabled', None)
        self.href = kwargs.get('href', None)
        self.position = kwargs.get('position', None)
        self.style = kwargs.get('style', None)
        self.text = kwargs.get('text', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'credits'

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, renders the credits text. Defaults to ``True``.

        :returns: Flag enabling or disabling credits.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def href(self) -> Optional[str]:
        """The URL for the credits label. Defaults to
        ``'https://www.highcharts.com?credits'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises ValueError: if not a URL, path-like string, or :obj:`None <python:None>`
        """
        return self._href

    @href.setter
    def href(self, value):
        self._href = validators.url(value, allow_empty = True)

    @property
    def position(self) -> Optional[Position]:
        """The position of the credits. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`Position` or :obj:`None <python:None>`
        """
        return self._position

    @position.setter
    @class_sensitive(Position)
    def position(self, value):
        self._position = value

    @property
    def style(self) -> Optional[CreditStyleOptions]:
        """CSS Styles for the credits label.

        Defaults to:

        .. code-block:: python

          {
              'color': '#999999',
              'cursor': 'pointer',
              'fontSize': '9px'
          }

        :rtype: :class:`CreditStyleOptions` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    @class_sensitive(CreditStyleOptions)
    def style(self, value):
        self._style = value

    @property
    def text(self) -> Optional[str]:
        """The text for the credits label. Defaults to
        ``'Highcharts'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._text

    @text.setter
    def text(self, value):
        self._text = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'enabled': as_dict.get('enabled', None),
            'href': as_dict.get('href', None),
            'position': as_dict.get('position', None),
            'style': as_dict.get('style', None),
            'text': as_dict.get('text', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'enabled': self.enabled,
            'href': self.href,
            'position': self.position,
            'style': self.style,
            'text': self.text,
        }

        return untrimmed



from typing import Optional

from highcharts_core.options.plot_options.series import SeriesOptions


class PolygonOptions(SeriesOptions):
    """General options to apply to all Polygon series types.

    A polygon series can be used to draw any freeform shape in the cartesian
    coordinate system. A fill is applied with the :meth:`PolygonOptions.color`
    setting, and stroke is applied through :meth:`PolygonOptions.line_width` and
    :meth:`PolygonOptions.line_color`.

    .. figure:: ../../../_static/polygon-example.png
      :alt: Polygon Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._track_by_area = None

        self.track_by_area = kwargs.get('track_by_area', None)

        super().__init__(**kwargs)

    @property
    def track_by_area(self) -> Optional[bool]:
        """When ``True``, the whole area should respond to mouseover tooltips and other
        mouse or touch events. When ``False``, only the line responds to mouse/touch
        events. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>`
        """
        return self._track_by_area

    @track_by_area.setter
    def track_by_area(self, value):
        if value is None:
            self._track_by_area = None
        else:
            self._track_by_area = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'step': as_dict.get('step', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'color_axis': as_dict.get('colorAxis', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'negative_color': as_dict.get('negativeColor', None),
            'point_description_format': as_dict.get('pointDescriptionFormat', None),
            'point_interval': as_dict.get('pointInterval', None),            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'stacking': as_dict.get('stacking', None),

            'track_by_area': as_dict.get('trackByArea', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'trackByArea': self.track_by_area,
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.utility_classes.data_labels import DataLabel as cls
from highcharts_core.utility_classes.data_labels import NodeDataLabel as cls2
from highcharts_core.utility_classes.data_labels import SunburstDataLabel as cls3
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'align': 'center',
      'allow_overlap': True,
      'animation': {
          'defer': 5
      },
      'background_color': {
          'linearGradient': {
              'x1': 0.123,
              'x2': 0.234,
              'y1': 0.345,
              'y2': 0.456
          },
          'stops': [
              [0.12, '#999'],
              [0.34, '#fff']
          ]
      },
      'border_color': '#999999',
      'border_radius': 24,
      'border_width': 1,
      'class_name': 'some-class-name',
      'color': '#000000',
      'crop': True,
      'defer': False,
      'enabled': True,
      'filter': {
          'operator': '>=',
          'property': 'some_property',
          'value': 123
      },
      'format': 'some format',
      'formatter': """function() { return true; }""",
      'inside': True,
      'null_format': 'some format',
      'null_formatter': """function() { return true; }""",
      'overflow': 'none',
      'padding': 12,
      'position': 'center',
      'rotation': 0,
      'shadow': False,
      'shape': 'rect',
      'style': 'style goes here',
      'use_html': False,
      'vertical_align': 'top',
      'x': 10,
      'y': 20,
      'z': 0
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_DataLabel__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_DataLabel__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_DataLabel_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_DataLabel_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/data_labels/01.js', False, None),

    ('utility_classes/data_labels/error-01.js', False, (errors.HighchartsValueError,
                                                        errors.HighchartsParseError,
                                                        JSONDecodeError,
                                                        TypeError,
                                                        ValueError)),

    ('utility_classes/data_labels/01.js', True, None),

    ('utility_classes/data_labels/error-01.js', True, (errors.HighchartsValueError,
                                                       errors.HighchartsParseError,
                                                       JSONDecodeError,
                                                       TypeError,
                                                       ValueError)),
])
def test_DataLabel_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)


STANDARD_PARAMS_2 = [
    ({}, None),
    ({
      'align': 'center',
      'allow_overlap': True,
      'animation': {
          'defer': 5
      },
      'background_color': {
          'linearGradient': {
              'x1': 0.123,
              'x2': 0.234,
              'y1': 0.345,
              'y2': 0.456
          },
          'stops': [
              [0.12, '#999'],
              [0.34, '#fff']
          ]
      },
      'border_color': '#999999',
      'border_radius': 24,
      'border_width': 1,
      'class_name': 'some-class-name',
      'color': '#000000',
      'crop': True,
      'defer': False,
      'enabled': True,
      'filter': {
          'operator': '>=',
          'property': 'some_property',
          'value': 123
      },
      'format': 'some format',
      'formatter': """function() { return true; }""",
      'inside': True,
      'null_format': 'some format',
      'null_formatter': """function() { return true; }""",
      'overflow': 'none',
      'padding': 12,
      'position': 'center',
      'rotation': 0,
      'shadow': False,
      'shape': 'rect',
      'style': 'style goes here',
      'use_html': False,
      'vertical_align': 'top',
      'x': 10,
      'y': 20,
      'z': 0
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_NodeDataLabel__init__(kwargs, error):
    Class__init__(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_NodeDataLabel__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_NodeDataLabel_from_dict(kwargs, error):
    Class_from_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_NodeDataLabel_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/data_labels/02.js', False, None),

    ('utility_classes/data_labels/error-03.js', False, (errors.HighchartsValueError,
                                                        errors.HighchartsParseError,
                                                        JSONDecodeError,
                                                        TypeError,
                                                        ValueError)),

    ('utility_classes/data_labels/02.js', True, None),

    ('utility_classes/data_labels/error-03.js', True, (errors.HighchartsValueError,
                                                       errors.HighchartsParseError,
                                                       JSONDecodeError,
                                                       TypeError,
                                                       ValueError)),
])
def test_NodeDataLabel_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls2, input_files, filename, as_file, error)


STANDARD_PARAMS_3 = [
    ({}, None),
    ({
      'align': 'center',
      'allow_overlap': True,
      'animation': {
          'defer': 5
      },
      'background_color': {
          'linearGradient': {
              'x1': 0.123,
              'x2': 0.234,
              'y1': 0.345,
              'y2': 0.456
          },
          'stops': [
              [0.12, '#999'],
              [0.34, '#fff']
          ]
      },
      'border_color': '#999999',
      'border_radius': 24,
      'border_width': 1,
      'class_name': 'some-class-name',
      'color': '#000000',
      'crop': True,
      'defer': False,
      'enabled': True,
      'filter': {
          'operator': '>=',
          'property': 'some_property',
          'value': 123
      },
      'format': 'some format',
      'formatter': """function() { return true; }""",
      'inside': True,
      'overflow': 'none',
      'padding': 12,
      'position': 'center',
      'rotation': 0,
      'rotation_mode': 'circular',
      'shadow': False,
      'shape': 'rect',
      'style': 'style goes here',
      'use_html': False,
      'vertical_align': 'top',
      'x': 10,
      'y': 20,
      'z': 0
    }, None),
    
    ({
      'align': 'center',
      'allow_overlap': True,
      'animation': {
          'defer': 5
      },
      'background_color': {
          'linearGradient': {
              'x1': 0.123,
              'x2': 0.234,
              'y1': 0.345,
              'y2': 0.456
          },
          'stops': [
              [0.12, '#999'],
              [0.34, '#fff']
          ]
      },
      'border_color': '#999999',
      'border_radius': 24,
      'border_width': 1,
      'class_name': 'some-class-name',
      'color': '#000000',
      'crop': True,
      'defer': False,
      'enabled': True,
      'filter': {
          'operator': '>=',
          'property': 'some_property',
          'value': 123
      },
      'format': 'some format',
      'formatter': """function() { return true; }""",
      'inside': True,
      'overflow': 'none',
      'padding': 12,
      'position': 'center',
      'rotation': 0,
      'rotation_mode': 'invalid-value',
      'shadow': False,
      'shape': 'rect',
      'style': 'style goes here',
      'use_html': False,
      'vertical_align': 'top',
      'x': 10,
      'y': 20,
      'z': 0
    }, errors.HighchartsValueError),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_3)
def test_SunburstDataLabel__init__(kwargs, error):
    Class__init__(cls3, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_3)
def test_SunburstDataLabel__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls3, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_3)
def test_SunburstDataLabel_from_dict(kwargs, error):
    Class_from_dict(cls3, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_3)
def test_SunburstDataLabel_to_dict(kwargs, error):
    Class_to_dict(cls3, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/data_labels/03.js', False, None),

    ('utility_classes/data_labels/error-03.js', False, (errors.HighchartsValueError,
                                                        errors.HighchartsParseError,
                                                        JSONDecodeError,
                                                        TypeError,
                                                        ValueError)),

    ('utility_classes/data_labels/03.js', True, None),

    ('utility_classes/data_labels/error-03.js', True, (errors.HighchartsValueError,
                                                       errors.HighchartsParseError,
                                                       JSONDecodeError,
                                                       TypeError,
                                                       ValueError)),
])
def test_SunburstDataLabel_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls3, input_files, filename, as_file, error)



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class CSVAnnotationOptions(HighchartsMeta):
    """Options for annotations in the export-data table."""

    def __init__(self, **kwargs):
        self._item_delimiter = None
        self._join = None

        self.item_delimiter = kwargs.get('item_delimiter', None)
        self.join = kwargs.get('join', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'exporting.csv.annotations'

    @property
    def item_delimiter(self) -> Optional[str]:
        """The way to mark the separator for annotations combined in one export-data table
        cell. Defaults to ``'; '``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._item_delimiter

    @item_delimiter.setter
    def item_delimiter(self, value):
        self._item_delimiter = validators.string(value, allow_empty = True)

    @property
    def join(self) -> Optional[bool]:
        """If ``True``, then when several labels are assigned to a specific point, they
        will be displayed in one field in the table.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._join

    @join.setter
    def join(self, value):
        if value is None:
            self._join = None
        else:
            self._join = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'item_delimiter': as_dict.get('itemDelimiter', None),
            'join': as_dict.get('join', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'itemDelimiter': self.item_delimiter,
            'join': self.join
        }

        return untrimmed


class ExportingCSV(HighchartsMeta):
    """Options for exporting data to CSV or Microsoft Excel, or displaying the data in
    a HTML table or a JavaScript structure.

    This module adds data export options to the export menu and provides JavaScript
    functions like ``Chart.getCSV()``, ``Chart.getTable()``, ``Chart.getDataRows()``,
    and ``Chart.viewData()``.

    .. warning::

      The XLS converter is limited and only creates a HTML string that is passed for
      download, which works but creates a warning before opening. The workaround for
      this is to use a third party XLSX converter.

    """

    def __init__(self, **kwargs):
        self._annotations = None
        self._column_header_formatter = None
        self._date_format = None
        self._decimal_point = None
        self._item_delimiter = None
        self._line_delimiter = None

        self.annotations = kwargs.get('annotations', None)
        self.column_header_formatter = kwargs.get('column_header_formatter', None)
        self.date_format = kwargs.get('date_format', None)
        self.decimal_point = kwargs.get('decimal_point', None)
        self.item_delimiter = kwargs.get('item_delimiter', None)
        self.line_delimiter = kwargs.get('line_delimiter', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'exporting.csv'

    @property
    def annotations(self) -> Optional[CSVAnnotationOptions]:
        """Options for annotations in the export-data table.

        :rtype: :class:`CSVAnnotationOption`
        """
        return self._annotations

    @annotations.setter
    @class_sensitive(CSVAnnotationOptions)
    def annotations(self, value):
        self._annotations = value

    @property
    def column_header_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript formatter callback function for the column headers.

        The parameters received by the callback function are:

          * ``item`` - The series or axis object
          * ``key`` - The point key, for example y or z
          * ``keyLength`` - The amount of value keys for this item, for example a range
            series has the keys low and high so the key length is 2.

        If :class:`Exporting.use_multi_level_headers` is ``True``, the JavaScript function
        should by default return an object with ``columnTitle`` and
        ``topLevelColumnTitle`` for each key. Columns with the same
        ``topLevelColumnTitle`` have their titles merged into a single cell with
        ``colspan`` for table/Excel export.

        If :class:`Exporting.use_multi_level_headers` is ``False``, or for CSV export, it
        should return the series name, followed by the key if there is more than one key.

        For the axis, it returns the axis title or "Category" or "DateTime" by default.

        To use Highcharts' automatically-suggested header, the callback function should
        return ``False``.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._column_header_formatter

    @column_header_formatter.setter
    @class_sensitive(CallbackFunction)
    def column_header_formatter(self, value):
        self._column_header_formatter = value

    @property
    def date_format(self) -> Optional[str]:
        """The date format to apply to exported dates on a datetime axis. Defaults to
        ``'%Y-%m-%d %H:%M:%S'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._date_format

    @date_format.setter
    def date_format(self, value):
        self._date_format = validators.string(value, allow_empty = True)

    @property
    def decimal_point(self) -> Optional[str | constants.EnforcedNullType]:
        """Decimal point to use for exported CSV. Defaults to the same as the browser
        locale, typically ``'.'`` (English) or ``','`` (German, French, etc).

        Returns either a string value, or an :class:`EnforcedNull <EnforcedNullType>`
        instance.

        :rtype: :class:`str <python:str>` or :class:`EnforcedNullType` or
          :obj:`None <python:None>`
        """
        return self._decimal_point

    @decimal_point.setter
    def decimal_point(self, value):
        if isinstance(value, constants.EnforcedNullType):
            self._decimal_point = constants.EnforcedNull
        else:
            self._decimal_point = validators.string(value, allow_empty = True)

    @property
    def item_delimiter(self) -> Optional[str | constants.EnforcedNullType]:
        """The item delimiter in the exported data.

        Use ``';'`` for direct exporting to Excel. If
        :class:`EnforcedNull <constants.EnforcedNullType>`, defaults to a best guess based
        on the browser locale. If the locale decimal point is ``','``, defaults to
        ``';'``. Otherwise, the delimiter defaults to ``','``.

        Returns either a string value, or an :class:`EnforcedNull <EnforcedNullType>`
        instance.

        :rtype: :class:`str <python:str>` or :class:`EnforcedNullType` or
          :obj:`None <python:None>`
        """
        return self._item_delimiter

    @item_delimiter.setter
    def item_delimiter(self, value):
        if isinstance(value, constants.EnforcedNullType):
            self._item_delimiter = constants.EnforcedNull
        else:
            self._item_delimiter = validators.string(value, allow_empty = True)

    @property
    def line_delimiter(self) -> Optional[str]:
        """The line delimiter in the exported data. Defaults to a new line.

        Returns either a string value, or an :class:`EnforcedNull <EnforcedNullType>`
        instance.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._line_delimiter

    @line_delimiter.setter
    def line_delimiter(self, value):
        self._line_delimiter = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'annotations': as_dict.get('annotations', None),
            'column_header_formatter': as_dict.get('columnHeaderFormatter', None),
            'date_format': as_dict.get('dateFormat', None),
            'decimal_point': as_dict.get('decimalPoint', None),
            'item_delimiter': as_dict.get('itemDelimiter', None),
            'line_delimiter': as_dict.get('lineDelimiter', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'annotations': self.annotations,
            'columnHeaderFormatter': self.column_header_formatter,
            'dateFormat': self.date_format,
            'decimalPoint': self.decimal_point,
            'itemDelimiter': self.item_delimiter,
            'lineDelimiter': self.line_delimiter
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.events import BreadcrumbEvents
from highcharts_core.utility_classes.position import Position
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class Separator(HighchartsMeta):
    """Configuration object for the Breadcrumbs Separator."""

    def __init__(self, **kwargs):
        self._style = None
        self._text = None

        self.style = kwargs.get('style', None)
        self.text = kwargs.get('text', None)

    @property
    def style(self) -> Optional[dict]:
        """CSS styles for the Breadcrumb Separator.

        :rtype: :class:`dict <python:dict>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.dict(value)

    @property
    def text(self) -> Optional[str]:
        """The text to use as the separator. Defaults to
        ``'/'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._text

    @text.setter
    def text(self, value):
        self._text = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'style': as_dict.get('style', None),
            'text': as_dict.get('text', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'style': self.style,
            'text': self.text
        }

        return untrimmed


class BreadcrumbOptions(HighchartsMeta):
    """Configuration for the breadcrumbs, the navigation at the top leading the way
    up through the drilldown levels."""

    def __init__(self, **kwargs):
        self._button_spacing = None
        self._button_theme = None
        self._events = None
        self._floating = None
        self._format = None
        self._formatter = None
        self._position = None
        self._relative_to = None
        self._rtl = None
        self._separator = None
        self._show_full_path = None
        self._style = None
        self._use_html = None
        self._z_index = None

        self.button_spacing = kwargs.get('button_spacing', None)
        self.button_theme = kwargs.get('button_theme', None)
        self.events = kwargs.get('events', None)
        self.floating = kwargs.get('floating', None)
        self.format = kwargs.get('format', None)
        self.formatter = kwargs.get('formatter', None)
        self.position = kwargs.get('position', None)
        self.relative_to = kwargs.get('relative_to', None)
        self.rtl = kwargs.get('rtl', None)
        self.separator = kwargs.get('separator', None)
        self.show_full_path = kwargs.get('show_full_path', None)
        self.style = kwargs.get('style', None)
        self.use_html = kwargs.get('use_html', None)
        self.z_index = kwargs.get('z_index', None)

    @property
    def button_spacing(self) -> Optional[int | float | Decimal]:
        """The default padding expressed in pixels for each button and separator in each
        direction.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._button_spacing

    @button_spacing.setter
    def button_spacing(self, value):
        self._button_spacing = validators.numeric(value, allow_empty = True)

    @property
    def button_theme(self) -> Optional[dict]:
        """A collection of attributes for the breadcrumb buttons.

        The object takes SVG attributes like ``fill``, ``stroke``, ``stroke-width`` or
        ``r``, the border radius.

        The theme also supports ``style``, a collection of CSS properties for the text.
        The object can also be extended with states, so you can set presentational options
        for ``hover``, ``select``, or ``disabled`` button states.

        :rtype: :class:`dict <python:dict>`
        """
        return self._theme

    @button_theme.setter
    def button_theme(self, value):
        self._theme = validators.dict(value, allow_empty = True)

    @property
    def events(self) -> Optional[BreadcrumbEvents]:
        """Definition of JavaScript event listeners to apply to the breadcrumbs.

        :rtype: :class:`ChartEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(BreadcrumbEvents)
    def events(self, value):
        self._events = value

    @property
    def floating(self) -> Optional[bool]:
        """If ``True`, sets the breadcrumbs to floating. When the breadcrumbs are
        floating, the plot area will not move to make space for it. Defaults to ``False``.

        .. warning::

          This property will not work when positioned in the middle.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._floating

    @floating.setter
    def floating(self, value):
        if value is None:
            self._floating = None
        else:
            self._floating = bool(value)

    @property
    def format(self) -> Optional[str]:
        """A format string for the breadcrumbs button. Variables are enclosed by curly
        brackets. Available values are passed in the declared point options.

        :returns: The format string to apply to the breadcrumbs.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._format

    @format.setter
    def format(self, value):
        self._format = validators.string(value, allow_empty = True)

    @property
    def formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format the breadcrumb text.

        :returns: A JavaScript callback function.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._formatter

    @formatter.setter
    @class_sensitive(CallbackFunction)
    def formatter(self, value):
        self._formatter = value

    @property
    def position(self) -> Optional[Position]:
        """The position of the button row.

        :rtype: :class:`Position` or :obj:`None <python:None>`
        """
        return self._position

    @position.setter
    @class_sensitive(Position)
    def position(self, value):
        self._position = value

    @property
    def relative_to(self) -> Optional[str]:
        """What frame the button placement should be related to. Defaults to
        ``'plotBox'``.

        Accepts:

          * ``'plot'``
          * ``'chart'``
          * ``'plotBox'``
          * ``'spacingBox'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._relative_to

    @relative_to.setter
    def relative_to(self, value):
        if not value:
            self._relative_to = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['plot', 'chart', 'plotBox', 'spacingBox']:
                raise errors.HighchartsValueError(f'relative_to accepts "plot", "chart", '
                                                  f'"plotBox", "spacingBox", or None. '
                                                  f'Received: {value}')
            self._relative_to = value

    @property
    def rtl(self) -> Optional[bool]:
        """If ``True`, reverses the order of the buttons (commonly used for Arabic or
        Hebrew). Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._rtl

    @rtl.setter
    def rtl(self, value):
        if value is None:
            self._rtl = None
        else:
            self._rtl = bool(value)

    @property
    def separator(self) -> Optional[Separator]:
        """Configuration for the breadcrumb separator.

        :rtype: :class:`Separator` or :obj:`None <python:None>`
        """
        return self._separator

    @separator.setter
    @class_sensitive(Separator)
    def separator(self, value):
        self._separator = value

    @property
    def show_full_path(self) -> Optional[bool]:
        """If ``True``, shows the full breadcrumb path. If ``False``, shows only a single
        breadcrumb button. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_full_path

    @show_full_path.setter
    def show_full_path(self, value):
        if value is None:
            self._show_full_path = None
        else:
            self._show_full_path = bool(value)

    @property
    def style(self) -> Optional[dict]:
        """CSS styles to apply to all breadcrumbs.

        :rtype: :class:`dict <python:dict>` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.dict(value, allow_empty = True)

    @property
    def use_html(self) -> Optional[bool]:
        """If ``True``, will use HTML to render the breadcrumbs. If ``False``, will
        use SVG or WebGL as applicable.

        Defaults to ``False``.

        :returns: Flag indicating whether to render the breadcrumbs using HTML.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._use_html

    @use_html.setter
    def use_html(self, value):
        if value is None:
            self._use_html = None
        else:
            self._use_html = bool(value)

    @property
    def z_index(self) -> Optional[int]:
        """The Z-Index for the breadcrumbs. Defaults to ``7``.

        :returns: The z-index for the annotation.
        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._z_index

    @z_index.setter
    def z_index(self, value):
        self._z_index = validators.integer(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'button_spacing': as_dict.get('buttonSpacing', None),
            'button_theme': as_dict.get('buttonTheme', None),
            'events': as_dict.get('events', None),
            'floating': as_dict.get('floating', None),
            'format': as_dict.get('format', None),
            'formatter': as_dict.get('formatter', None),
            'position': as_dict.get('position', None),
            'relative_to': as_dict.get('relativeTo', None),
            'rtl': as_dict.get('rtl', None),
            'separator': as_dict.get('separator', None),
            'show_full_path': as_dict.get('showFullPath', None),
            'style': as_dict.get('style', None),
            'use_html': as_dict.get('useHTML', None),
            'z_index': as_dict.get('zIndex', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'buttonSpacing': self.button_spacing,
            'buttonTheme': self.button_theme,
            'events': self.events,
            'floating': self.floating,
            'format': self.format,
            'formatter': self.formatter,
            'position': self.position,
            'relativeTo': self.relative_to,
            'rtl': self.rtl,
            'separator': self.separator,
            'showFullPath': self.show_full_path,
            'style': self.style,
            'useHTML': self.use_html,
            'zIndex': self.z_index
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.utility_classes.events import ChartEvents as cls
from highcharts_core.utility_classes.events import BreadcrumbEvents as cls2
from highcharts_core.utility_classes.events import NavigationEvents as cls3
from highcharts_core.utility_classes.events import PointEvents as cls4
from highcharts_core.utility_classes.events import SeriesEvents as cls5
from highcharts_core.utility_classes.events import ClusterEvents as cls6
from highcharts_core.utility_classes.events import AxisEvents as cls7
from highcharts_core.utility_classes.events import MouseEvents as cls8

from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'add_series': """function(event) { return true;}""",
      'after_print': """function(event) {return true;}""",
      'click': """function(event) { return true; }""",
      'selection': """function(event) { return true; }"""
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_ChartEvents__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_ChartEvents__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_ChartEvents_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_ChartEvents_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/events/01.js', False, None),

    ('utility_classes/events/error-00.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),
    ('utility_classes/events/error-01.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),

    ('utility_classes/events/01.js', True, None),

    ('utility_classes/events/error-00.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
    ('utility_classes/events/error-01.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
])
def test_ChartEvents_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)


STANDARD_PARAMS_2 = [
    ({}, None),
    ({
      'click': """function(event) { return true; }"""
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_BreadcrumbEvents__init__(kwargs, error):
    Class__init__(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_BreadcrumbEvents__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_BreadcrumbEvents_from_dict(kwargs, error):
    Class_from_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_BreadcrumbEvents_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/events/02.js', False, None),

    ('utility_classes/events/error-00.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),
    ('utility_classes/events/error-02.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),

    ('utility_classes/events/02.js', True, None),

    ('utility_classes/events/error-00.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
    ('utility_classes/events/error-02.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
])
def test_BreadcrumbEvents_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls2, input_files, filename, as_file, error)


STANDARD_PARAMS_3 = [
    ({}, None),
    ({
      'close_popup': """function (event) { return true; }""",
      'select_button': """function (event) {return true;}""",
      'show_popup': """function(event) {return true;}"""
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_3)
def test_NavigationEvents__init__(kwargs, error):
    Class__init__(cls3, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_3)
def test_NavigationEvents__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls3, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_3)
def test_NavigationEvents_from_dict(kwargs, error):
    Class_from_dict(cls3, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_3)
def test_NavigationEvents_to_dict(kwargs, error):
    Class_to_dict(cls3, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/events/03.js', False, None),

    ('utility_classes/events/error-00.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),
    ('utility_classes/events/error-03.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),

    ('utility_classes/events/03.js', True, None),

    ('utility_classes/events/error-00.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
    ('utility_classes/events/error-03.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
])
def test_NavigationEvents_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls3, input_files, filename, as_file, error)


STANDARD_PARAMS_4 = [
    ({}, None),
    ({
      'click': """function(event) { return true; }""",
      'drag': """function(event) { return true; }""",
      'drop': """function(event) { return true; }""",
      'mouse_out': """function(event) { return true; }"""
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_4)
def test_PointEvents__init__(kwargs, error):
    Class__init__(cls4, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_4)
def test_PointEvents__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls4, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_4)
def test_PointEvents_from_dict(kwargs, error):
    Class_from_dict(cls4, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_4)
def test_PointEvents_to_dict(kwargs, error):
    Class_to_dict(cls4, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/events/04.js', False, None),

    ('utility_classes/events/error-00.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),
    ('utility_classes/events/error-04.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),

    ('utility_classes/events/04.js', True, None),

    ('utility_classes/events/error-00.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
    ('utility_classes/events/error-04.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
])
def test_PointEvents_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls4, input_files, filename, as_file, error)


STANDARD_PARAMS_5 = [
    ({}, None),
    ({
      'after_animate': """function(event) { return true; }""",
      'click': """function(event) { return true; }""",
      'hide': """function(event) { return true; }""",
      'mouse_out': """function(event) { return true; }""",
      'show': """function(event) { return true; }"""
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_5)
def test_SeriesEvents__init__(kwargs, error):
    Class__init__(cls5, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_5)
def test_SeriesEvents__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls5, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_5)
def test_SeriesEvents_from_dict(kwargs, error):
    Class_from_dict(cls5, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_5)
def test_SeriesEvents_to_dict(kwargs, error):
    Class_to_dict(cls5, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/events/05.js', False, None),

    ('utility_classes/events/error-00.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),
    ('utility_classes/events/error-05.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),

    ('utility_classes/events/05.js', True, None),

    ('utility_classes/events/error-00.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
    ('utility_classes/events/error-05.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
])
def test_SeriesEvents_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls5, input_files, filename, as_file, error)


STANDARD_PARAMS_6 = [
    ({}, None),
    ({
      'drill_to_cluster': """function(event) { return true; }"""
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_6)
def test_ClusterEvents__init__(kwargs, error):
    Class__init__(cls6, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_6)
def test_ClusterEvents__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls6, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_6)
def test_ClusterEvents_from_dict(kwargs, error):
    Class_from_dict(cls6, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_6)
def test_ClusterEvents_to_dict(kwargs, error):
    Class_to_dict(cls6, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/events/06.js', False, None),

    ('utility_classes/events/error-00.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),
    ('utility_classes/events/error-06.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),

    ('utility_classes/events/06.js', True, None),

    ('utility_classes/events/error-00.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
    ('utility_classes/events/error-06.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
])
def test_ClusterEvents_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls6, input_files, filename, as_file, error)


STANDARD_PARAMS_7 = [
    ({}, None),
    ({
      'after_breaks': """function(event) { return true; }""",
      'after_set_extremes': """function(event) { return true; }""",
      'point_break': """function(event) { return true; }""",
      'set_extremes': """function(event) { return true; }"""
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_7)
def test_AxisEvents__init__(kwargs, error):
    Class__init__(cls7, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_7)
def test_AxisEvents__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls7, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_7)
def test_AxisEvents_from_dict(kwargs, error):
    Class_from_dict(cls7, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_7)
def test_AxisEvents_to_dict(kwargs, error):
    Class_to_dict(cls7, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/events/07.js', False, None),

    ('utility_classes/events/error-00.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),
    ('utility_classes/events/error-07.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),

    ('utility_classes/events/07.js', True, None),

    ('utility_classes/events/error-00.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
    ('utility_classes/events/error-07.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
])
def test_AxisEvents_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls7, input_files, filename, as_file, error)


STANDARD_PARAMS_8 = [
    ({}, None),
    ({
      'click': """function(event) { return true; }""",
      'mousemove': """function(event) { return true; }""",
      'mouseout': """function(event) { return true; }""",
      'mouseover': """function(event) { return true; }"""
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_8)
def test_MouseEvents__init__(kwargs, error):
    Class__init__(cls8, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_8)
def test_MouseEvents__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls8, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_8)
def test_MouseEvents_from_dict(kwargs, error):
    Class_from_dict(cls8, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_8)
def test_MouseEvents_to_dict(kwargs, error):
    Class_to_dict(cls8, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/events/08.js', False, None),

    ('utility_classes/events/error-00.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),
    ('utility_classes/events/error-08.js', False, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),

    ('utility_classes/events/08.js', True, None),

    ('utility_classes/events/error-00.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
    ('utility_classes/events/error-08.js', True, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),
])
def test_MouseEvents_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls8, input_files, filename, as_file, error)
